from datetime import timedelta
from typing import Optional, Tuple

from . import types as command_types
from opentrons.protocol_api.tasks import Task


def comment(msg: str) -> command_types.CommentCommand:
    text = msg
    return {"name": command_types.COMMENT, "payload": {"text": text}}


def delay(
    seconds: float, minutes: float, msg: Optional[str] = None
) -> command_types.DelayCommand:
    td = timedelta(minutes=minutes, seconds=seconds)
    actual_min, actual_sec = divmod(td.total_seconds(), 60)

    text = f"Delaying for {int(actual_min)} minutes and {round(actual_sec, 3)} seconds"

    if msg:
        text = f"{text}. {msg}"

    return {
        "name": command_types.DELAY,
        "payload": {"minutes": actual_min, "seconds": actual_sec, "text": text},
    }


def pause(msg: Optional[str] = None) -> command_types.PauseCommand:
    text = "Pausing robot operation"
    if msg:
        text = text + ": {}".format(msg)
    return {
        "name": command_types.PAUSE,
        "payload": {
            "text": text,
            "userMessage": msg,
        },
    }


def resume() -> command_types.ResumeCommand:
    return {
        "name": command_types.RESUME,
        "payload": {"text": "Resuming robot operation"},
    }


def move_labware(text: str) -> command_types.MoveLabwareCommand:
    return {
        "name": command_types.MOVE_LABWARE,
        "payload": {"text": text},
    }


def capture_image(
    resolution: Optional[Tuple[int, int]],
    zoom: Optional[float],
    contrast: Optional[float],
    brightness: Optional[float],
    saturation: Optional[float],
) -> command_types.CaptureImageCommand:
    text = "Capturing an image"
    if resolution:
        text += f" with resolution {resolution[0]}x{resolution[1]}"
    if zoom:
        text += f" zoom of {zoom}X"
    if contrast:
        text += f" contrast of {contrast}%"
    if brightness:
        text += f" brightness of {brightness}%"
    if saturation:
        text += f" saturation of {saturation}%"
    text += "."
    return {
        "name": command_types.CAPTURE_IMAGE,
        "payload": {
            "text": text,
            "resolution": resolution,
            "zoom": zoom,
            "contrast": contrast,
            "brightness": brightness,
            "saturation": saturation,
        },
    }


def wait_for_tasks(tasks: list[Task]) -> command_types.WaitForTasksCommand:
    task_ids = [task.created_at.strftime("%Y-%m-%d %H:%M:%S") for task in tasks]
    msg = f"Waiting for tasks that started at: {task_ids}."
    return {
        "name": command_types.WAIT_FOR_TASKS,
        "payload": {"text": msg},
    }


def create_timer(seconds: float) -> command_types.CreateTimerCommand:
    return {
        "name": command_types.CREATE_TIMER,
        "payload": {
            "text": f"Creating background timer for {seconds} seconds.",
            "time": seconds,
        },
    }



"""Test aspirate commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    PipetteOverpressureError,
    StallOrCollisionDetectedError,
)

from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine import (
    DeckPoint,
    LiquidHandlingWellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine.commands.aspirate import (
    AspirateImplementation,
    AspirateParams,
    AspirateResult,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.execution import (
    MovementHandler,
    PipettingHandler,
)
from opentrons.protocol_engine.notes import CommandNoteAdder
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    AspiratedFluid,
    CurrentWell,
    FluidKind,
    LabwareWellId,
    WellLocation,
)
from opentrons.types import Point


@pytest.fixture
def subject(
    state_view: StateView,
    hardware_api: HardwareControlAPI,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    mock_command_note_adder: CommandNoteAdder,
    model_utils: ModelUtils,
) -> AspirateImplementation:
    """Get the implementation subject."""
    return AspirateImplementation(
        pipetting=pipetting,
        state_view=state_view,
        movement=movement,
        hardware_api=hardware_api,
        command_note_adder=mock_command_note_adder,
        model_utils=model_utils,
    )


async def test_aspirate_implementation_no_prep(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    mock_command_note_adder: CommandNoteAdder,
) -> None:
    """An Aspirate should have an execution implementation without preparing to aspirate."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )
    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )

    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(True)

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["covered-well-1", "covered-well-2"])

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-50,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            command_note_adder=mock_command_note_adder,
            correction_volume=0.0,
        ),
    ).then_return(50)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=AspirateResult(volume=50, position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id=labware_id,
                well_names=["covered-well-1", "covered-well-2"],
                volume_added=-100,
            ),
            pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                pipette_id=pipette_id,
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=50),
            ),
        ),
    )


async def test_aspirate_implementation_with_prep(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    mock_command_note_adder: CommandNoteAdder,
    subject: AspirateImplementation,
) -> None:
    """An Aspirate should have an execution implementation with preparing to aspirate."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )
    volume = 50
    flow_rate = 1.23
    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=location,
        volume=volume,
        flowRate=flow_rate,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        False
    )

    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(False)
    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["covered-well-1", "covered-well-2"])

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=WellLocation(origin=WellOrigin.TOP),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point())

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=location,
            current_well=CurrentWell(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
            ),
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-volume,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=volume,
            flow_rate=flow_rate,
            command_note_adder=mock_command_note_adder,
            correction_volume=0.0,
        ),
    ).then_return(volume)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=AspirateResult(volume=50, position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id=labware_id,
                well_names=["covered-well-1", "covered-well-2"],
                volume_added=-100,
            ),
            pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                pipette_id=pipette_id,
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=50),
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id=pipette_id, ready_to_aspirate=True
            ),
        ),
    )


async def test_aspirate_raises_volume_error(
    decoy: Decoy,
    pipetting: PipettingHandler,
    movement: MovementHandler,
    mock_command_note_adder: CommandNoteAdder,
    state_view: StateView,
    subject: AspirateImplementation,
) -> None:
    """Should raise an assertion error for volume larger than working volume."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )
    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )

    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(True)
    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["covered-well-1", "covered-well-2"])

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-50,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(1, 2, 3))

    decoy.when(
        await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            command_note_adder=mock_command_note_adder,
            correction_volume=0.0,
        )
    ).then_raise(AssertionError("blah blah"))

    with pytest.raises(AssertionError):
        await subject.execute(params)


async def test_overpressure_error(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    model_utils: ModelUtils,
    mock_command_note_adder: CommandNoteAdder,
    state_view: StateView,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["covered-well-1", "covered-well-2"])

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )
    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(True)

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-50,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(position)

    decoy.when(
        await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            command_note_adder=mock_command_note_adder,
            correction_volume=0.0,
        ),
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=OverpressureError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (position.x, position.y, position.z)},
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
                new_deck_point=DeckPoint(x=position.x, y=position.y, z=position.z),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id=labware_id,
                well_names=["covered-well-1", "covered-well-2"],
                volume_added=update_types.CLEAR,
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id=pipette_id
            ),
        ),
    )


async def test_aspirate_implementation_meniscus(
    decoy: Decoy,
    state_view: StateView,
    hardware_api: HardwareControlAPI,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    mock_command_note_adder: CommandNoteAdder,
) -> None:
    """Aspirate should update WellVolumeOffset when called with WellOrigin.MENISCUS."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    location = LiquidHandlingWellLocation(
        origin=WellOrigin.MENISCUS,
        offset=WellOffset(x=0, y=0, z=-1),
        volumeOffset="operationVolume",
    )
    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["covered-well-1", "covered-well-2"])

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )
    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(True)

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-50,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        await pipetting.aspirate_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            command_note_adder=mock_command_note_adder,
            correction_volume=0,
        ),
    ).then_return(50)

    result = await subject.execute(params)

    assert result == SuccessData(
        public=AspirateResult(volume=50, position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id=labware_id,
                well_names=["covered-well-1", "covered-well-2"],
                volume_added=-100,
            ),
            pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                pipette_id=pipette_id,
                fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=50),
            ),
        ),
    )


async def test_stall_during_final_movement(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should propagate a stall error that happens when moving to the final position."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)
    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        True
    )
    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(True)

    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=-50,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_raise(StallOrCollisionDetectedError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(params)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )


async def test_stall_during_preparation(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should propagate a stall error that happens during the prepare-to-aspirate part."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        False
    )
    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(False)

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=WellLocation(origin=WellOrigin.TOP),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_raise(StallOrCollisionDetectedError())
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(params)
    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id, createdAt=error_timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
        ),
        state_update_if_false_positive=update_types.StateUpdate(),
    )


async def test_overpressure_during_preparation(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: AspirateImplementation,
    state_view: StateView,
    model_utils: ModelUtils,
) -> None:
    """It should propagate an overpressure error that happens during the prepare-to-aspirate part."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    params = AspirateParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(pipetting.get_is_ready_to_aspirate(pipette_id=pipette_id)).then_return(
        False
    )
    decoy.when(state_view.pipettes.get_ready_to_aspirate(pipette_id)).then_return(False)

    retry_location = Point(1, 2, 3)
    decoy.when(
        state_view.geometry.get_well_position(
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            operation_volume=-params.volume,
            pipette_id=pipette_id,
        )
    ).then_return(retry_location)

    prep_location = Point(4, 5, 6)
    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=WellLocation(origin=WellOrigin.TOP),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(prep_location)

    decoy.when(await pipetting.prepare_for_aspirate(pipette_id)).then_raise(  # type: ignore[func-returns-value]
        PipetteOverpressureError()
    )
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(params)
    assert result == DefinedErrorData(
        public=OverpressureError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={
                "retryLocation": (retry_location.x, retry_location.y, retry_location.z)
            },
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
                new_deck_point=DeckPoint(
                    x=prep_location.x, y=prep_location.y, z=prep_location.z
                ),
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id=pipette_id
            ),
        ),
        state_update_if_false_positive=update_types.StateUpdate(),
    )



"""Tests for the InstrumentContext public interface."""

import inspect
from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict

from . import versions_at_or_above, versions_at_or_below, versions_between
from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
    Labware,
    TemperatureModuleContext,
    Well,
)
from opentrons.protocol_api._liquid import Liquid
from opentrons.protocol_api.core import well_grid
from opentrons.protocol_api.core.common import (
    LabwareCore,
    ModuleCore,
    ProtocolCore,
    WellCore,
)
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError, UnsupportedAPIError
from opentrons.types import Point


@pytest.fixture(autouse=True)
def _mock_well_grid_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out the well grid module."""
    for name, func in inspect.getmembers(well_grid, inspect.isfunction):
        monkeypatch.setattr(well_grid, name, decoy.mock(func=func))


@pytest.fixture
def mock_labware_core(decoy: Decoy) -> LabwareCore:
    """Get a mock labware implementation core."""
    return decoy.mock(cls=LabwareCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_map_core(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock map core."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def api_version() -> APIVersion:
    """Get the API version to test at."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
    api_version: APIVersion,
) -> Labware:
    """Get a Labware test subject with its dependencies mocked out."""
    decoy.when(mock_labware_core.get_well_columns()).then_return([])
    decoy.when(well_grid.create([])).then_return(
        well_grid.WellGrid(columns_by_name={}, rows_by_name={})
    )
    return Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 13)])
def test_api_version(api_version: APIVersion, subject: Labware) -> None:
    """It should have an api_version property."""
    assert subject.api_version == api_version


@pytest.mark.parametrize("api_version", [APIVersion(2, 12)])
def test_api_version_clamped(subject: Labware) -> None:
    """It should not allow the API version to go any lower than 2.13."""
    assert subject.api_version == APIVersion(2, 13)


def test_is_tiprack(
    decoy: Decoy, mock_labware_core: LabwareCore, subject: Labware
) -> None:
    """It should report if it's a tip rack."""
    decoy.when(mock_labware_core.is_tip_rack()).then_return(True)
    assert subject.is_tiprack is True

    decoy.when(mock_labware_core.is_tip_rack()).then_return(False)
    assert subject.is_tiprack is False


def test_is_adapter(
    decoy: Decoy, mock_labware_core: LabwareCore, subject: Labware
) -> None:
    """It should report if it's an adapter."""
    decoy.when(mock_labware_core.is_adapter()).then_return(True)
    assert subject.is_adapter is True

    decoy.when(mock_labware_core.is_adapter()).then_return(False)
    assert subject.is_adapter is False


def test_load_labware(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
    api_version: APIVersion,
    subject: Labware,
) -> None:
    """It should load a labware by the load parameters."""
    new_mock_core = decoy.mock(cls=LabwareCore)
    decoy.when(
        mock_protocol_core.load_labware(
            load_name="labware-name",
            label="a label",
            namespace="a-namespace",
            version=123,
            location=mock_labware_core,
        )
    ).then_return(new_mock_core)
    decoy.when(new_mock_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        name="labware-name",
        label="a label",
        namespace="a-namespace",
        version=123,
    )

    assert isinstance(result, Labware)
    assert result.api_version == api_version
    decoy.verify(mock_map_core.add(new_mock_core, result), times=1)


def test_load_labware_from_definition(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
    api_version: APIVersion,
    subject: Labware,
) -> None:
    """It should be able to load a labware from a definition dictionary."""
    new_mock_core = decoy.mock(cls=LabwareCore)

    labware_definition_dict = cast(LabwareDefDict, {"labwareDef": True})
    labware_load_params = LabwareLoadParams("you", "are", 1337)

    decoy.when(
        mock_protocol_core.add_labware_definition(labware_definition_dict)
    ).then_return(labware_load_params)

    decoy.when(new_mock_core.get_well_columns()).then_return([])

    decoy.when(
        mock_protocol_core.load_labware(
            namespace="you",
            load_name="are",
            version=1337,
            label="a label",
            location=mock_labware_core,
        )
    ).then_return(new_mock_core)

    result = subject.load_labware_from_definition(
        definition=labware_definition_dict,
        label="a label",
    )

    assert isinstance(result, Labware)
    assert result.api_version == api_version
    decoy.verify(mock_map_core.add(new_mock_core, result), times=1)


def test_wells(
    decoy: Decoy,
    api_version: APIVersion,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
    subject: Labware,
) -> None:
    """It should return a list of wells."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)
    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )

    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_1.get_max_volume()).then_return(100)
    decoy.when(mock_well_core_2.get_name()).then_return("B1")
    decoy.when(mock_well_core_1.get_max_volume()).then_return(1000)

    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)

    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    result = subject.wells()
    result_a1 = result[0]
    result_b1 = result[1]

    assert len(result) == 2
    assert isinstance(result_a1, Well)
    assert result_a1.well_name == "A1"
    assert isinstance(result_b1, Well)
    assert result_b1.well_name == "B1"

    assert subject.wells_by_name() == {"A1": result_a1, "B1": result_b1}

    assert subject.rows() == [[result_a1], [result_b1]]
    assert subject.rows_by_name() == {"A": [result_a1], "B": [result_b1]}
    assert subject.columns() == [[result_a1, result_b1]]
    assert subject.columns_by_name() == {"1": [result_a1, result_b1]}


def test_reset_tips(
    decoy: Decoy, mock_labware_core: LabwareCore, subject: Labware
) -> None:
    """It should reset and tip state."""
    subject.reset()
    decoy.verify(mock_labware_core.reset_tips(), times=1)


def test_set_empty(
    decoy: Decoy, mock_labware_core: LabwareCore, subject: Labware
) -> None:
    """It should empty the tip state."""
    subject.set_empty()
    decoy.verify(mock_labware_core.set_empty(), times=1)


def test_parent_slot(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
) -> None:
    """Should get the labware's parent slot name or None."""
    decoy.when(mock_protocol_core.get_labware_location(mock_labware_core)).then_return(
        "bloop"
    )

    assert subject.parent == "bloop"


def test_parent_module_context(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """Should get labware parent module context."""
    mock_module_core = decoy.mock(cls=ModuleCore)
    mock_temp_module_context = decoy.mock(cls=TemperatureModuleContext)

    decoy.when(mock_protocol_core.get_labware_location(mock_labware_core)).then_return(
        mock_module_core
    )

    decoy.when(mock_map_core.get(mock_module_core)).then_return(
        mock_temp_module_context
    )

    assert subject.parent == mock_temp_module_context


def test_parent_labware(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """Should get labware's parent labware."""
    mock_parent_labware_core = decoy.mock(cls=LabwareCore)
    mock_labware = decoy.mock(cls=Labware)

    decoy.when(mock_protocol_core.get_labware_location(mock_labware_core)).then_return(
        mock_parent_labware_core
    )

    decoy.when(mock_map_core.get(mock_parent_labware_core)).then_return(mock_labware)

    assert subject.parent == mock_labware


def test_child(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should get the labware sitting on top of the labware."""
    mock_child_labware_core = decoy.mock(cls=LabwareCore)
    mock_labware = decoy.mock(cls=Labware)

    decoy.when(
        mock_protocol_core.get_labware_on_labware(mock_labware_core)
    ).then_return(mock_child_labware_core)
    decoy.when(mock_map_core.get(mock_child_labware_core)).then_return(mock_labware)

    assert subject.child == mock_labware


@pytest.mark.parametrize("api_version", versions_at_or_below(APIVersion(2, 13)))
def test_set_offset_succeeds_on_low_api_version(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
) -> None:
    """It should pass the offset to the core, on low API versions."""
    subject.set_offset(1, 2, 3)
    decoy.verify(mock_labware_core.set_calibration(Point(1, 2, 3)))


@pytest.mark.parametrize(
    "api_version",
    versions_between(
        low_inclusive_bound=APIVersion(2, 14), high_inclusive_bound=APIVersion(2, 17)
    ),
)
def test_set_offset_raises_on_intermediate_api_version(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
) -> None:
    """It should raise an error, on high API versions."""
    with pytest.raises(APIVersionError):
        subject.set_offset(1, 2, 3)


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 18)))
def test_set_offset_succeeds_on_high_api_version(
    decoy: Decoy, subject: Labware, mock_labware_core: LabwareCore
) -> None:
    """It should not raise an API version error on the most recent versions."""
    subject.set_offset(1, 2, 3)
    decoy.verify(mock_labware_core.set_calibration(Point(1, 2, 3)))


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 14)))
def test_separate_calibration_raises_on_high_api_version(
    decoy: Decoy,
    subject: Labware,
    mock_labware_core: LabwareCore,
) -> None:
    """It should raise an error, on high API versions."""
    with pytest.raises(UnsupportedAPIError):
        subject.separate_calibration


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_liquid_handles_valid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should load volumes for list of wells."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)
    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")

    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)

    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    mock_liquid = decoy.mock(cls=Liquid)

    subject.load_liquid(["A1", subject["B1"]], 10, mock_liquid)
    decoy.verify(
        mock_labware_core.load_liquid(
            {
                "A1": 10,
                "B1": 10,
            },
            mock_liquid,
        )
    )


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_liquid_rejects_invalid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should require valid load inputs."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)

    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")
    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)
    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )

    core_2 = decoy.mock(cls=LabwareCore)
    mock_well_core_3 = decoy.mock(cls=WellCore)
    grid_2 = well_grid.WellGrid(
        columns_by_name={"1": ["A1"]}, rows_by_name={"A": ["A1"]}
    )
    decoy.when(mock_well_core_3.get_name()).then_return("A1")
    decoy.when(core_2.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(core_2.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(core_2.get_well_core("B1")).then_return(mock_well_core_2)

    decoy.when(well_grid.create([["A1"]])).then_return(grid_2)
    other_labware = Labware(
        core=core_2,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    mock_liquid = decoy.mock(cls=Liquid)
    with pytest.raises(KeyError):
        subject.load_liquid(["A1", "C1"], 10, mock_liquid)

    with pytest.raises(KeyError):
        subject.load_liquid([subject["A1"], other_labware["A1"]], 10, mock_liquid)

    with pytest.raises(TypeError):
        subject.load_liquid([2], 10, mock_liquid)  # type: ignore[list-item]

    with pytest.raises(TypeError):
        subject.load_liquid(["A1"], "A1", mock_liquid)  # type: ignore[arg-type]
    mock_liquid = decoy.mock(cls=Liquid)

    subject.load_liquid(["A1", subject["B1"]], 10, mock_liquid)
    decoy.verify(
        mock_labware_core.load_liquid(
            {
                "A1": 10,
                "B1": 10,
            },
            mock_liquid,
        )
    )


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_liquid_by_well_handles_valid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should load liquids of different volumes in different wells."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)
    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")

    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)
    decoy.when(mock_well_core_2.get_display_name()).then_return("well 2")
    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    mock_liquid = decoy.mock(cls=Liquid)

    subject.load_liquid_by_well({"A1": 10, subject["B1"]: 11}, mock_liquid)
    decoy.verify(
        mock_labware_core.load_liquid(
            {
                "A1": 10,
                "B1": 11,
            },
            mock_liquid,
        )
    )


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_liquid_by_well_rejects_invalid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should require valid well specs."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)

    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")
    decoy.when(mock_well_core_1.get_display_name()).then_return("well 1")
    decoy.when(mock_well_core_2.get_display_name()).then_return("well 2")
    decoy.when(mock_well_core_1.get_top(z_offset=0.0)).then_return(Point(4, 5, 6))
    decoy.when(mock_well_core_1.get_top(z_offset=0.0)).then_return(Point(7, 8, 9))
    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)
    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    core_2 = decoy.mock(cls=LabwareCore)
    mock_well_core_3 = decoy.mock(cls=WellCore)
    decoy.when(mock_well_core_3.get_display_name()).then_return("well 3")
    grid_2 = well_grid.WellGrid(
        columns_by_name={"1": ["A1"]}, rows_by_name={"A": ["A1"]}
    )
    decoy.when(mock_well_core_3.get_name()).then_return("A1")
    decoy.when(core_2.get_well_columns()).then_return([["A1"]])
    decoy.when(core_2.get_well_core("A1")).then_return(mock_well_core_3)
    decoy.when(mock_well_core_3.get_top(z_offset=0.0)).then_return(Point(1, 2, 3))

    decoy.when(well_grid.create([["A1"]])).then_return(grid_2)
    other_labware = Labware(
        core=core_2,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )

    mock_liquid = decoy.mock(cls=Liquid)
    with pytest.raises(KeyError):
        subject.load_liquid_by_well({"A1": 10, "C1": 11}, mock_liquid)

    with pytest.raises(KeyError):
        subject.load_liquid_by_well(
            {subject["A1"]: 10, other_labware["A1"]: 11}, mock_liquid
        )

    with pytest.raises(TypeError):
        subject.load_liquid_by_well({2: 10}, mock_liquid)  # type: ignore[dict-item]

    with pytest.raises(TypeError):
        subject.load_liquid_by_well({"A1": "A3"}, mock_liquid)  # type: ignore[dict-item]


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_empty_handles_valid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should load lists of wells as empty."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)
    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")

    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)

    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )

    subject.load_empty(["A1", subject["B1"]])
    decoy.verify(mock_labware_core.load_empty(["A1", "B1"]))


@pytest.mark.parametrize("api_version", versions_at_or_above(APIVersion(2, 22)))
def test_load_empty_rejects_invalid_inputs(
    decoy: Decoy,
    mock_labware_core: LabwareCore,
    api_version: APIVersion,
    mock_protocol_core: ProtocolCore,
    mock_map_core: LoadedCoreMap,
) -> None:
    """It should require valid well specs."""
    mock_well_core_1 = decoy.mock(cls=WellCore)
    mock_well_core_2 = decoy.mock(cls=WellCore)

    grid = well_grid.WellGrid(
        columns_by_name={"1": ["A1", "B1"]},
        rows_by_name={"A": ["A1"], "B": ["B1"]},
    )
    decoy.when(mock_well_core_1.get_name()).then_return("A1")
    decoy.when(mock_well_core_2.get_name()).then_return("B1")
    decoy.when(mock_labware_core.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(mock_labware_core.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(mock_labware_core.get_well_core("B1")).then_return(mock_well_core_2)
    decoy.when(well_grid.create([["A1", "B1"]])).then_return(grid)
    subject = Labware(
        core=mock_labware_core,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )

    core_2 = decoy.mock(cls=LabwareCore)
    mock_well_core_3 = decoy.mock(cls=WellCore)
    grid_2 = well_grid.WellGrid(
        columns_by_name={"1": ["A1"]}, rows_by_name={"A": ["A1"]}
    )
    decoy.when(mock_well_core_3.get_name()).then_return("A1")
    decoy.when(core_2.get_well_columns()).then_return([["A1", "B1"]])
    decoy.when(core_2.get_well_core("A1")).then_return(mock_well_core_1)
    decoy.when(core_2.get_well_core("B1")).then_return(mock_well_core_2)

    decoy.when(well_grid.create([["A1"]])).then_return(grid_2)
    other_labware = Labware(
        core=core_2,
        api_version=api_version,
        protocol_core=mock_protocol_core,
        core_map=mock_map_core,
    )
    with pytest.raises(KeyError):
        subject.load_empty(["A1", "C1"])

    with pytest.raises(KeyError):
        subject.load_empty([subject["A1"], other_labware["A1"]])

    with pytest.raises(TypeError):
        subject.load_empty([2])  # type: ignore[list-item]



import asyncio
import re
from typing import Optional

from .abstract import AbstractVacuumModuleDriver
from .errors import VacuumModuleErrorCodes
from .types import (
    GCODE,
    HardwareRevision,
    LEDColor,
    LEDPattern,
    PressureState,
    PumpState,
    VacuumModuleInfo,
    VentState,
)
from opentrons.drivers.asyncio.communication import AsyncResponseSerialConnection

VM_BAUDRATE = 115200
DEFAULT_VM_TIMEOUT = 5
VM_ACK = "OK\n"
VM_ERROR_KEYWORD = "err"
VM_ASYNC_ERROR_ACK = "async"
DEFAULT_COMMAND_RETRIES = 2
GCODE_ROUNDING_PRECISION = 2

# LED animation range values
MIN_DURATION_MS = 25  # 25ms
MAX_DURATION_MS = 10000  # 10s
MAX_REPS = 10

MAX_PUMP_RPM = 3500
MAX_PUMP_DUTY = 90
MAX_RAMP_RATE = -10.0  # mbar/s
MAX_PRESSURE_MBAR = -1013.25


class VacuumModuleDriver(AbstractVacuumModuleDriver):
    """Driver for Opentrons Vacuum Module."""

    @classmethod
    def parse_device_info(cls, response: str) -> VacuumModuleInfo:
        """Parse vacuum module info."""
        _RE = re.compile(
            f"^{GCODE.GET_DEVICE_INFO} FW:(?P<fw>\\S+) HW:Opentrons-vacuum-module-(?P<hw>\\S+) SerialNo:(?P<sn>\\S+)$"
        )
        m = _RE.match(response)
        if not m:
            raise ValueError(f"Incorrect Response for device info: {response}")
        return VacuumModuleInfo(
            m.group("fw"), HardwareRevision(m.group("hw")), m.group("sn")
        )

    @classmethod
    def parse_reset_reason(cls, response: str) -> int:
        """Parse the reset reason"""
        _RE = re.compile(rf"^{GCODE.GET_RESET_REASON} R:(?P<R>\d)$")
        match = _RE.match(response)
        if not match:
            raise ValueError(f"Incorrect Response for reset reason: {response}")
        return int(match.group("R"))

    @classmethod
    def parse_get_pressure_state(cls, response: str) -> PressureState:
        """Parse the get pressure state."""
        pattern = r"T:(?P<T>-?\d.+) C:(?P<C>-?\d.+) A:(?P<A>\d.+) B:(?P<B>\d.+) H:(?P<H>\d.+) E:(?P<E>\d) V:(?P<V>\d)"
        _RE = re.compile(rf"^{GCODE.GET_PRESSURE_STATE} {pattern}$")
        match = _RE.match(response)
        if not match:
            raise ValueError(f"Incorrect Response for get pressure state: {response}")
        return PressureState(
            float(match.group("T")),
            float(match.group("C")),
            float(match.group("A")),
            float(match.group("B")),
            float(match.group("H")),
            bool(int(match.group("E"))),
            VentState(int(match.group("V"))),
        )

    @classmethod
    def parse_get_pump_state(cls, response: str) -> PumpState:
        """Parse the get pump state."""
        pattern = r"T:(?P<T>\d.+) R:(?P<R>\d.+) A:(?P<A>\d+) D:(?P<D>\d+) E:(?P<E>\d) M:(?P<M>\d)"
        _RE = re.compile(rf"^{GCODE.GET_PUMP_STATE} {pattern}$")
        match = _RE.match(response)
        if not match:
            raise ValueError(f"Incorrect Response for get pump state: {response}")
        return PumpState(
            float(match.group("T")),
            float(match.group("R")),
            int(match.group("A")),
            int(match.group("D")),
            bool(int(match.group("E"))),
            bool(int(match.group("M"))),
        )

    @classmethod
    async def create(
        cls, port: str, loop: Optional[asyncio.AbstractEventLoop]
    ) -> "VacuumModuleDriver":
        """Create a Vacuum Module driver."""
        connection = await AsyncResponseSerialConnection.create(
            port=port,
            baud_rate=VM_BAUDRATE,
            timeout=DEFAULT_VM_TIMEOUT,
            number_of_retries=DEFAULT_COMMAND_RETRIES,
            ack=VM_ACK,
            loop=loop,
            error_keyword=VM_ERROR_KEYWORD,
            async_error_ack=VM_ASYNC_ERROR_ACK,
            reset_buffer_before_write=True,
            error_codes=VacuumModuleErrorCodes,
        )
        return cls(connection)

    def __init__(self, connection: AsyncResponseSerialConnection) -> None:
        """
        Constructor

        Args:
            connection: connection to the vacuum module
        """
        self._connection = connection

    async def connect(self) -> None:
        """Connect to vacuum module."""
        await self._connection.open()

    async def disconnect(self) -> None:
        """Disconnect from vacuum module."""
        await self._connection.close()

    async def is_connected(self) -> bool:
        """Check connection to vacuum module."""
        return await self._connection.is_open()

    def reset_serial_buffers(self) -> None:
        """Reset the input and output serial buffers."""
        self._connection._serial.reset_input_buffer()
        self._connection._serial.reset_output_buffer()

    async def get_device_info(self) -> VacuumModuleInfo:
        """Get Device Info."""
        response = await self._connection.send_command(
            GCODE.GET_DEVICE_INFO.build_command()
        )
        device_info = self.parse_device_info(response)
        reason_resp = await self._connection.send_command(
            GCODE.GET_RESET_REASON.build_command()
        )
        reason = self.parse_reset_reason(reason_resp)
        device_info.rr = reason
        return device_info

    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        command = GCODE.ENTER_BOOTLOADER.build_command()
        await self._connection.send_dfu_command(command)
        await self._connection.close()

    async def set_serial_number(self, sn: str) -> None:
        """Set Serial Number."""
        if not re.match(r"^VM[\w]{1}[\d]{2}[\d]{8}[\d]+$", sn):
            raise ValueError(
                f"Invalid serial number: ({sn}) expected format: VMA1020250119001"
            )
        resp = await self._connection.send_command(
            GCODE.SET_SERIAL_NUMBER.build_command().add_element(sn)
        )
        if not re.match(rf"^{GCODE.SET_SERIAL_NUMBER}$", resp):
            raise ValueError(f"Incorrect Response for set serial number: {resp}")

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Set LED Status bar color and pattern.

        :param power: Power of the LED (0-1.0), 0 is off, 1 is full power
        :param color: Color of the LED
        :param external: True if external LED, False if internal LED
        :param pattern: Animation pattern of the LED status bar
        :param duration: Animation duration in milliseconds (25-10000), 10s max
        :param reps: Number of times to repeat the animation (-1 - 10), -1 is forever.
        """
        power = max(0, min(power, 1.0))
        command = GCODE.SET_LED.build_command().add_float(
            "P", power, GCODE_ROUNDING_PRECISION
        )
        if color is not None:
            command.add_int("C", color.value)
        if external is not None:
            command.add_int("K", int(external))
        if pattern is not None:
            command.add_int("A", pattern.value)
        if duration is not None:
            duration = max(MIN_DURATION_MS, min(duration, MAX_DURATION_MS))
            command.add_int("D", duration)
        if reps is not None:
            command.add_int("R", max(-1, min(reps, MAX_REPS)))
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_LED}$", resp):
            raise ValueError(f"Incorrect Response for set led: {resp}")

    async def set_vacuum_state(
        self,
        enable_vacuum: bool,
        guage_pressure_mbar: Optional[float] = None,
        duration: Optional[int] = None,
        rate: Optional[float] = None,
        vent_after: Optional[bool] = None,
    ) -> None:
        """Engage or release the vacuum until a desired internal pressure is reached."""

        command = GCODE.SET_PRESSURE_STATE.build_command().add_int(
            "S", int(enable_vacuum)
        )

        if guage_pressure_mbar is not None:
            command.add_float(
                "P",
                min(max(guage_pressure_mbar, MAX_PRESSURE_MBAR), 0),
                GCODE_ROUNDING_PRECISION,
            )
        if duration is not None:
            command.add_int("D", duration)
        if rate is not None:
            command.add_float("R", min(max(rate, MAX_RAMP_RATE), 0))
        if vent_after is not None:
            command.add_int("V", int(vent_after))

        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_PRESSURE_STATE}$", resp):
            raise ValueError(f"Incorrect Response for set pressure state: {resp}")

    async def get_vacuum_state(self) -> PressureState:
        """Get the pressure state."""
        resp = await self._connection.send_command(
            GCODE.GET_PRESSURE_STATE.build_command()
        )
        return self.parse_get_pressure_state(resp)

    async def set_pump_state(
        self,
        start_pump: bool,
        target_rpm: Optional[int] = None,
        duty_cycle: Optional[int] = None,
    ) -> None:
        """Start or the stop the pump at a given rpm or duty cycle."""
        if target_rpm and duty_cycle:
            raise ValueError(
                "You cannot set the target rpm and duty cycle at the same time."
            )

        command = GCODE.SET_PUMP_STATE.build_command().add_int("S", int(start_pump))
        if target_rpm is not None:
            command.add_int("R", max(0, min(target_rpm, MAX_PUMP_RPM)))
        if duty_cycle is not None:
            command.add_int("D", max(0, min(duty_cycle, MAX_PUMP_DUTY)))
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_PUMP_STATE}$", resp):
            raise ValueError(f"Incorrect Response for set pump state: {resp}")

    async def get_pump_state(self) -> PumpState:
        """Get the pump state."""
        resp = await self._connection.send_command(GCODE.GET_PUMP_STATE.build_command())
        return self.parse_get_pump_state(resp)

    # turns off motor, then releases, takes a timeout for buffer between turn off and vent
    async def set_vent_state(self, state: bool) -> None:
        """Opens/Closes the vent, which release the vacuum in the module chamber."""

        command = GCODE.SET_VENT_STATE.build_command().add_int("V", int(state))
        resp = await self._connection.send_command(command)
        if not re.match(rf"^{GCODE.SET_VENT_STATE}$", resp):
            raise ValueError(f"Incorrect Response for set vent state: {resp}")



from typing import List

import pytest

from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore
from opentrons.protocols.api_support.tip_tracker import TipTracker


@pytest.fixture
def well_ordering() -> List[List[str]]:
    return [[f"{c}{r}" for c in (chr(65 + i) for i in range(8))] for r in range(1, 13)]


@pytest.fixture
def wells_by_column(well_ordering: List[List[str]]) -> List[List[LegacyWellCore]]:
    return [
        [
            LegacyWellCore(
                well_geometry=None,  # type: ignore[arg-type]
                display_name=well_name,
                has_tip=True,
                name=well_name,
            )
            for well_name in column
        ]
        for column in well_ordering
    ]


@pytest.fixture
def wells(wells_by_column: List[List[LegacyWellCore]]) -> List[LegacyWellCore]:
    return [well_impl for column in wells_by_column for well_impl in column]


@pytest.fixture
def tiptracker(wells_by_column: List[List[LegacyWellCore]]) -> TipTracker:
    return TipTracker(wells_by_column)


def test_use_tips(wells: List[LegacyWellCore], tiptracker: TipTracker) -> None:
    well_list = wells

    # Test only using one tip
    tiptracker.use_tips(well_list[0])
    assert not well_list[0].has_tip()
    for well in well_list[1:]:
        assert well.has_tip()

    # Test using a whole column
    tiptracker.use_tips(well_list[8], num_channels=8)
    for well in well_list[8:16]:
        assert not well.has_tip()
    assert well_list[7].has_tip()
    assert well_list[16].has_tip()

    # Test using a partial column from the top
    tiptracker.use_tips(well_list[16], num_channels=4)
    for well in well_list[16:20]:
        assert not well.has_tip()
    for well in well_list[20:24]:
        assert well.has_tip()

    # Test using a partial column where the number of tips that get picked up
    # is less than the number of channels (e.g.: an 8-channel pipette picking
    # up 4 tips from the bottom half of a column)
    tiptracker.use_tips(well_list[28], num_channels=4)
    for well in well_list[24:28]:
        assert well.has_tip()
    for well in well_list[28:32]:
        assert not well.has_tip()
    for well in well_list[32:]:
        assert well.has_tip()


def test_select_next_tip(wells: List[LegacyWellCore], tiptracker: TipTracker) -> None:
    well_list = wells

    next_one = tiptracker.next_tip()
    assert next_one == well_list[0]
    next_five = tiptracker.next_tip(5)
    assert next_five == well_list[0]
    next_eight = tiptracker.next_tip(8)
    assert next_eight == well_list[0]
    next_nine = tiptracker.next_tip(9)
    assert next_nine is None

    # A1 tip only has been used
    tiptracker.use_tips(well_list[0])

    next_one = tiptracker.next_tip()
    assert next_one == well_list[1]
    next_five = tiptracker.next_tip(5)
    assert next_five == well_list[1]
    next_eight = tiptracker.next_tip(8)
    assert next_eight == well_list[8]

    # 2nd column has also been used
    tiptracker.use_tips(well_list[8], num_channels=8)

    next_one = tiptracker.next_tip()
    assert next_one == well_list[1]
    next_five = tiptracker.next_tip(5)
    assert next_five == well_list[1]
    next_eight = tiptracker.next_tip(8)
    assert next_eight == well_list[16]

    # Bottom 4 tips of 1rd column are also used
    tiptracker.use_tips(well_list[4], num_channels=4)

    next_one = tiptracker.next_tip()
    assert next_one == well_list[1]
    next_three = tiptracker.next_tip(3)
    assert next_three == well_list[1]
    next_five = tiptracker.next_tip(5)
    assert next_five == well_list[16]
    next_eight = tiptracker.next_tip(8)
    assert next_eight == well_list[16]

    # you can reuse tips infinitely on api level 2.2
    tiptracker.use_tips(well_list[0])
    tiptracker.use_tips(well_list[0])


def test_use_tips_fail_if_full(
    wells: List[LegacyWellCore], tiptracker: TipTracker
) -> None:
    well_list = wells

    tiptracker.use_tips(well_list[0])
    with pytest.raises(AssertionError):
        tiptracker.use_tips(well_list[0], fail_if_full=True)


def test_previous_tip(wells: List[LegacyWellCore], tiptracker: TipTracker) -> None:
    # If all wells are used, we can't get a previous tip
    assert tiptracker.previous_tip() is None
    # If one well is empty, wherever it is, we can get a slot
    wells[5].set_has_tip(False)
    assert tiptracker.previous_tip() is wells[5]
    # But not if we ask for more slots than are available
    assert tiptracker.previous_tip(2) is None
    wells[7].set_has_tip(False)
    # And those available wells have to be contiguous
    assert tiptracker.previous_tip(2) is None
    # But if they are, we're good
    wells[6].set_has_tip(False)
    assert tiptracker.previous_tip(3) is wells[5]


def test_return_tips(wells: List[LegacyWellCore], tiptracker: TipTracker) -> None:
    # If all wells are used, we get an error if we try to return
    with pytest.raises(AssertionError):
        tiptracker.return_tips(wells[0])
    # If we have space where we specify, everything is OK
    wells[0].set_has_tip(False)
    tiptracker.return_tips(wells[0])
    assert wells[0].has_tip()
    # We have to have enough space
    wells[0].set_has_tip(False)
    with pytest.raises(AssertionError):
        tiptracker.return_tips(wells[0], 2)
    # But we can drop stuff off the end of a column
    wells[7].set_has_tip(False)
    wells[8].set_has_tip(False)
    tiptracker.return_tips(wells[7], 2)
    assert wells[7].has_tip()
    # But we won't wrap around
    assert not wells[8].has_tip()



import os
import random
import tempfile
from pathlib import Path

import pytest

from opentrons import config
from opentrons.system import wifi


@pytest.fixture()
def wifi_keys_tempdir(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Path:
    wifi_keys_dir = tmp_path / "wifi_keys"
    wifi_keys_dir.mkdir()
    monkeypatch.setitem(config.CONFIG, "wifi_keys_dir", wifi_keys_dir)

    return wifi_keys_dir


def test_check_eap_config(wifi_keys_tempdir: Path) -> None:
    wifi_key_id = "88188cafcf"
    os.mkdir(os.path.join(wifi_keys_tempdir, wifi_key_id))
    with open(os.path.join(wifi_keys_tempdir, wifi_key_id, "test.pem"), "w") as f:
        f.write("what a terrible key")
    # Bad eap types should fail
    with pytest.raises(wifi.ConfigureArgsError):
        wifi.eap_check_config({"eapType": "afaosdasd"})
    # Valid (if short) arguments should work
    wifi.eap_check_config(
        {
            "eapType": "peap/eap-mschapv2",
            "identity": "test@hi.com",
            "password": "passwd",
        }
    )
    # Extra args should fail
    with pytest.raises(wifi.ConfigureArgsError):
        wifi.eap_check_config(
            {
                "eapType": "tls",
                "identity": "test@example.com",
                "privateKey": wifi_key_id,
                "clientCert": wifi_key_id,
                "phase2CaCertf": "foo",
            }
        )
    # Filenames should be rewritten
    rewritten = wifi.eap_check_config(
        {
            "eapType": "ttls/eap-md5",
            "identity": "hello@example.com",
            "password": "hi",
            "caCert": wifi_key_id,
        }
    )
    assert rewritten["caCert"] == os.path.join(
        wifi_keys_tempdir, wifi_key_id, "test.pem"
    )
    # A config should be returned with the same keys
    config = {
        "eapType": "ttls/eap-tls",
        "identity": "test@hello.com",
        "phase2ClientCert": wifi_key_id,
        "phase2PrivateKey": wifi_key_id,
    }
    out = wifi.eap_check_config(config)
    for key in config.keys():
        assert key in out


def test_eap_check_option() -> None:
    # Required arguments that are not specified should raise
    with pytest.raises(wifi.ConfigureArgsError):
        wifi._eap_check_option_ok(
            {
                "name": "test-opt",
                # TODO(mc, 2021-09-12): typechecking expects string
                "required": True,  # type: ignore[dict-item]
                "displayName": "Test Option",
            },
            {"eapType": "test"},
        )
    # Non-required arguments that are not specified should not raise
    wifi._eap_check_option_ok(
        {
            "name": "test-1",
            # TODO(mc, 2021-09-12): typechecking expects string
            "required": False,  # type: ignore[dict-item]
            "type": "string",
            "displayName": "Test Option",
        },
        {"eapType": "test"},
    )

    # Check type mismatch detection pos and neg
    with pytest.raises(wifi.ConfigureArgsError):
        wifi._eap_check_option_ok(
            {
                "name": "identity",
                "displayName": "Username",
                # TODO(mc, 2021-09-12): typechecking expects string
                "required": True,  # type: ignore[dict-item]
                "type": "string",
            },
            {"identity": 2, "eapType": "test"},
        )
    wifi._eap_check_option_ok(
        {
            "name": "identity",
            # TODO(mc, 2021-09-12): typechecking expects string
            "required": True,  # type: ignore[dict-item]
            "displayName": "Username",
            "type": "string",
        },
        {"identity": "hi", "eapType": "test"},
    )
    with pytest.raises(wifi.ConfigureArgsError):
        wifi._eap_check_option_ok(
            {
                "name": "password",
                # TODO(mc, 2021-09-12): typechecking expects string
                "required": True,  # type: ignore[dict-item]
                "displayName": "Password",
                "type": "password",
            },
            {"password": [2, 3], "eapType": "test"},
        )
    wifi._eap_check_option_ok(
        {
            "name": "password",
            # TODO(mc, 2021-09-12): typechecking expects string
            "required": True,  # type: ignore[dict-item]
            "displayName": "password",
            "type": "password",
        },
        {"password": "secret", "eapType": "test"},
    )
    with pytest.raises(wifi.ConfigureArgsError):
        wifi._eap_check_option_ok(
            {
                "name": "phase2CaCert",
                "displayName": "some file who cares",
                # TODO(mc, 2021-09-12): typechecking expects string
                "required": True,  # type: ignore[dict-item]
                "type": "file",
            },
            {"phase2CaCert": 2, "eapType": "test"},
        )
    wifi._eap_check_option_ok(
        {
            "name": "phase2CaCert",
            # TODO(mc, 2021-09-12): typechecking expects string
            "required": True,  # type: ignore[dict-item]
            "displayName": "hello",
            "type": "file",
        },
        {"phase2CaCert": "82141cceaf", "eapType": "test"},
    )


async def test_list_keys(wifi_keys_tempdir: Path) -> None:
    dummy_names = ["ad12d1df199bc912", "cbdda8124128cf", "812410990c5412"]
    for dn in dummy_names:
        os.mkdir(os.path.join(wifi_keys_tempdir, dn))
        with open(os.path.join(wifi_keys_tempdir, dn, "test.pem"), "w") as file:
            file.write("hi")

    keys = list(wifi.list_keys())
    assert len(keys) == 3
    for dn in dummy_names:
        for keyfile in keys:
            if keyfile.directory == dn:
                assert keyfile.file == "test.pem"
                break
        else:
            raise KeyError(dn)


async def test_key_lifecycle(wifi_keys_tempdir: Path) -> None:
    with tempfile.TemporaryDirectory() as source_td:
        keys = list(wifi.list_keys())
        assert keys == []

        results = {}
        # We should be able to add multiple keys
        for fn in ["test1.pem", "test2.pem", "test3.pem"]:
            path = os.path.join(source_td, fn)
            with open(path, "w") as f:
                f.write(str(random.getrandbits(2048)))

            # TODO(mc, 2021-09-12): use pathlib
            with open(path, "rb") as f:
                add_response = wifi.add_key(fn, f.read())
                assert add_response.created is True
                assert add_response.key.file == fn
                results[fn] = add_response

        # We should not be able to upload a duplicate
        # TODO(mc, 2021-09-12): use pathlib
        with open(os.path.join(source_td, "test1.pem"), "rb") as f:
            add_response = wifi.add_key("test1.pem", f.read())
            assert add_response.created is False

        # We should be able to see them all
        list_resp = list(wifi.list_keys())
        assert len(list_resp) == 3
        for elem in list_resp:
            assert elem.directory in {r.key.directory for r in results.values()}

        for fn, data in results.items():
            del_resp = wifi.remove_key(data.key.directory)
            assert del_resp == fn

            del_list_resp = list(wifi.list_keys())
            assert data.key.directory not in {k.directory for k in del_list_resp}

        dup_del_resp = wifi.remove_key(results["test1.pem"].key.directory)
        assert dup_del_resp is None



import pytest

from opentrons.protocol_api.core.legacy.legacy_well_core import LegacyWellCore


@pytest.mark.parametrize(
    argnames=["name", "row", "col"],
    argvalues=[
        ["A1", "A", "1"],
        ["A10", "A", "10"],
        ["Z100", "Z", "100"],
        ["A0", "A", "0"],
        ["B0100", "B", "0100"],
    ],
)
def test_row_column(name: str, row: str, col: str) -> None:
    w = LegacyWellCore(
        well_geometry=None,  # type: ignore[arg-type]
        display_name=None,  # type: ignore[arg-type]
        has_tip=None,  # type: ignore[arg-type]
        name=name,
    )
    assert w.get_name() == name
    assert w.get_row_name() == row
    assert w.get_column_name() == col


@pytest.mark.parametrize(
    argnames=["name"],
    argvalues=[
        ["a1"],
        ["Aa1"],
        ["A 1"],
    ],
)
def test_row_column_fail(name: str) -> None:
    with pytest.raises(AssertionError, match=f"could not match '{name}'"):
        LegacyWellCore(
            well_geometry=None,  # type: ignore[arg-type]
            display_name=None,  # type: ignore[arg-type]
            has_tip=None,  # type: ignore[arg-type]
            name=name,
        )



"""Test labware validation."""

import pytest

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    LabwareDefinition2,
    LabwareRole,
    Parameters2,
    Vector3D,
)

from opentrons.protocol_engine.resources import labware_validation as subject


@pytest.mark.parametrize(
    ("definition", "expected_result"),
    [
        (
            LabwareDefinition2.model_construct(allowedRoles=[LabwareRole.labware]),  # type: ignore[call-arg]
            True,
        ),
        (
            LabwareDefinition2.model_construct(allowedRoles=[]),  # type: ignore[call-arg]
            True,
        ),
        (
            LabwareDefinition2.model_construct(allowedRoles=[LabwareRole.adapter]),  # type: ignore[call-arg]
            False,
        ),
    ],
)
def test_validate_definition_is_labware(
    definition: LabwareDefinition, expected_result: bool
) -> None:
    """It should validate if definition is defined as a labware."""
    assert subject.validate_definition_is_labware(definition) == expected_result


@pytest.mark.parametrize(
    ("definition", "expected_result"),
    [
        (
            LabwareDefinition2.model_construct(allowedRoles=[LabwareRole.adapter]),  # type: ignore[call-arg]
            True,
        ),
        (
            LabwareDefinition2.model_construct(allowedRoles=[]),  # type: ignore[call-arg]
            False,
        ),
        (
            LabwareDefinition2.model_construct(allowedRoles=[LabwareRole.labware]),  # type: ignore[call-arg]
            False,
        ),
    ],
)
def test_validate_definition_is_adapter(
    definition: LabwareDefinition, expected_result: bool
) -> None:
    """It should validate if definition is defined as an adapter."""
    assert subject.validate_definition_is_adapter(definition) == expected_result


@pytest.mark.parametrize(
    ("definition", "expected_result"),
    [
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                stackingOffsetWithLabware={"labware123": Vector3D(x=4, y=5, z=6)}
            ),
            True,
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                stackingOffsetWithLabware={"labwareXYZ": Vector3D(x=4, y=5, z=6)}
            ),
            False,
        ),
        (
            LabwareDefinition2.model_construct(stackingOffsetWithLabware={}),  # type: ignore[call-arg]
            False,
        ),
    ],
)
def test_validate_labware_can_be_stacked(
    definition: LabwareDefinition2, expected_result: bool
) -> None:
    """It should validate if definition allows it to stack on given labware."""
    assert (
        subject.validate_legacy_labware_can_be_stacked(definition, "labware123")
        == expected_result
    )


@pytest.mark.parametrize(
    ("definition", "expected_result"),
    [
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(quirks=None)  # type: ignore[call-arg]
            ),
            True,
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(quirks=["foo"])  # type: ignore[call-arg]
            ),
            True,
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                parameters=Parameters2.model_construct(quirks=["gripperIncompatible"])  # type: ignore[call-arg]
            ),
            False,
        ),
    ],
)
def test_validate_gripper_compatible(
    definition: LabwareDefinition, expected_result: bool
) -> None:
    """It should validate if definition is defined as an adapter."""
    assert subject.validate_gripper_compatible(definition) == expected_result
